##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
"""
Zopeish DOMish XML Node classes.

"""
import Acquisition
from ExtensionClass import Base
import string

# Node type codes
# ---------------

ELEMENT_NODE       = 1
ATTRIBUTE_NODE     = 2
TEXT_NODE          = 3
CDATA_SECTION_NODE = 4
ENTITY_REFERENCE_NODE = 5
ENTITY_NODE        = 6
PROCESSING_INSTRUCTION_NODE = 7
COMMENT_NODE       = 8
DOCUMENT_NODE      = 9
DOCUMENT_TYPE_NODE = 10
DOCUMENT_FRAGMENT_NODE = 11
NOTATION_NODE      = 12

# Exception codes
# ---------------

INDEX_SIZE_ERR              = 1
DOMSTRING_SIZE_ERR          = 2
HIERARCHY_REQUEST_ERR       = 3
WRONG_DOCUMENT_ERR          = 4
INVALID_CHARACTER_ERR       = 5
NO_DATA_ALLOWED_ERR         = 6
NO_MODIFICATION_ALLOWED_ERR = 7
NOT_FOUND_ERR               = 8
NOT_SUPPORTED_ERR           = 9
INUSE_ATTRIBUTE_ERR         = 10

# Exceptions
# ----------

class DOMException(Exception):
    pass
class IndexSizeException(DOMException):
    code = INDEX_SIZE_ERR
class DOMStringSizeException(DOMException):
    code = DOMSTRING_SIZE_ERR
class HierarchyRequestException(DOMException):
    code = HIERARCHY_REQUEST_ERR
class WrongDocumentException(DOMException):
    code = WRONG_DOCUMENT_ERR
class NoDataAllowedException(DOMException):
    code = NO_DATA_ALLOWED_ERR
class NoModificationAllowedException(DOMException):
    code = NO_MODIFICATION_ALLOWED_ERR
class NotFoundException(DOMException):
    code = NOT_FOUND_ERR
class NotSupportedException(DOMException):
    code = NOT_SUPPORTED_ERR
class InUseAttributeException(DOMException):
    code = INUSE_ATTRIBUTE_ERR

        
# Node classes
# ------------
             
class Node(Acquisition.Implicit):
    """Node astract base class"""
    
    # _child_map maps child ids to child element objects
    # _child_ids is a list of child element object ids in order
    # id is the Node id
    # name is the Node name
    # attributes is a property sheet of node attributes

    # note: _child_map and _child_ids might be replaced with 
    # higher performance map and set objects in Nodes with
    # large numbers of children. also methods that return
    # lists of nodes might be done with some kind of lazy
    # iteration, rather than computing the whole big list
    # at once.
    
    isXMLNode=1
    
    def __init__(self, name, attributes=None):
        self._name=name
        self._child_ids=[]
        self._child_map={}
    
    # private methods
    # ---------------
    
    def _make_id(self, parent=None):
        # create a unique id for this node
        if parent is None:
            id=0
        else:
            while 1:
                if hasattr(parent,'aq_parent') \
                        and hasattr(parent.aq_parent, 'isXMLNode') \
                        and parent.aq_parent.isXMLNode:
                    parent=parent.aq_parent
                else:
                    break
            if hasattr(parent, '_next_id'):
                id=parent._next_id
                parent._next_id=id + 1
            else:
                id=0
                parent._next_id=1
        self._id=id
        self.id=self._index2id(id)
        
    def _id2index(self,id):
        # convert Node id to index in _child_map
        return string.atoi(id[1:])
            
    def _index2id(self,index):
        # convert id in _child_map to Node id
        return 'e%d' % index
    
    def _append_child(self, child):
        # used by builder to construct DOM tree
        child._make_id(self)
        self._child_ids.append(child._id)
        self._child_map[child._id]=child
        
    def _clear(self):
        # used by builder to clear the node of children and attributes
        self._clear_attributes()
        self._child_ids=[]
        self._child_map.clear()
    
    def _clear_attributes(self):
        if hasattr(self, '_attributes'):
            self._attributes={}
    
    # sequence interface to children by index
    # mapping interface to childen by id
    # ---------------------------------------

    # TODO add __setitem__, __delitem__, __getslice__, __setslice__
    # append, insert, etc.
    
    def __getitem__(self, key):
        try: key=string.atoi(key)
        except: pass
        if type(key)==type(1):
            try:
                return self._child_map[self._child_ids[key]].__of__(self)
            except:
                raise IndexError('index out of range')
        try:
            key=self._id2index(key)
            return self._child_map[key].__of__(self)
        except:
            raise KeyError(key)       
                
    def __len__(self):
        return len(self._child_ids)

    # DOM Interface
    # -------------

    nodeType=None
    
    def getnodeType(self):
        "The type code of this node."
        return self.nodeType
        
    def getnodeValue(self):
        "The value of this node."
        return None
    
    def getnodeName(self):
        "The name of this node."
        return self._name
        
    def getattributes(self):
        """
        A Python dictionary containing the attributes of this node
        (if it is an Element) or None otherwise.
        """
        return None
        
    def getparentNode(self):
        """
        The parent of this node, or None if it doesn't exist.
        """
        try:
            if self.aq_parent.isXMLNode:
                return self.aq_parent
        except AttributeError:
            return None
                
    def getpreviousSibling(self):
        """
        The node immediately preceding this node.
        If there is no such node, this returns None. 
        """
        try:
            i=self.aq_parent._child_ids.index(self._id)
            if i-1 <= 0:
                return None
            return self.aq_parent[i-1]
        except:
            return None
    
    def getnextSibling(self):
        """
        The node immediately following this node.
        If there is no such node, this returns None. 
        """
        try:
            i=self.aq_parent._child_ids.index(self._id)
            if i+1 >= len(self.aq_parent):
                return None
            return self.aq_parent[i+1]
        except:
            return None
        
    def getownerDocument(self):
        """
        The Document object associated with this node.
        If there is no such node, this returns None.        
        """
        node=self
        while 1:
            if node.getnodeType()==DOCUMENT_NODE:
                return node
            elif hasattr(node,'aq_parent'):
                node=node.aq_parent
            else:
                return None
                
    def getchildNodes(self):
        """
        A Python list that contains all children of this node.
        If there are no children, this is an empty list.
        
        Note: This list is not live, i.e. changes to the list do
        not effect the children of this node.
        """
        return self.objectValues()    
        
    def getfirstChild(self):
        """
        The first child of this node.
        If there is no such node, this returns None.  
        """
        try:
            return self[0]
        except KeyError:
            return None
    
    def getlastChild(self):
        """
        The last child of this node.
        If there is no such node, this returns None.  
        """
        try:
            return self[-1]
        except KeyError:
            return None    
        
    def hasChildNodes(self):
        """
        true if the node has any children, false if the node has no children.
        """
        return len(self)
            
    # DOM write methods
    # -----------------
    
    def normalize(self):
        """
        Consolidate child Text nodes.
        """
        last=None
        remove_ids=[]
        for child in self:
            if child.getnodeType() == TEXT_NODE and last is not None \
                    and last.getnodeType() == TEXT_NODE:
                last._data=last._data + child._data
                remove_ids.append(child._id)
            else:
                last=child
                if last.getnodeType() != TEXT_NODE:
                    last.normalize()
        for id in remove_ids:
            self._child_ids.remove(id)
            del self._child_map[id]

    def insertBefore(self, newChild, refChild=None):
        """
        Inserts the node newChild before the existing child node refChild.
        If refChild is None, insert newChild at the end of the list of children.
        If the newChild is already in the tree, it is first removed. 
        """
        if refChild is None:
            self.appendChild(newChild)
            return
        try:
            i=self._child_ids.index(refChild._id)
        except IndexError:
            raise NotFoundException(refChild)
        if newChild.getparentNode():
            newChild.getparentNode().removeChild(newChild)
            newChild=newChild.aq_base
        self._child_ids.insert(i, newChild._id)
        self._child_map[newChild._id]=newChild
        
    def replaceChild(self, newChild, oldChild):
        """
        Replaces the child node oldChild with newChild in the list of children,
        and returns the oldChild node. If the newChild is already in the tree,
        it is first removed. 
        """       
        try:
            i=self._child_ids.index(oldChild._id)
        except IndexError:
            raise NotFoundException(oldChild)
        if newChild.getparentNode():
            newChild.getparentNode().removeChild(newChild)
            newChild=newChild.aq_base
        self._child_ids[i]=newChild._id
        del self._child_map[oldChild._id]
        self._child_map[newChild._id]=newChild
        return oldChild
        
    def removeChild(self, oldChild):
        """
        Removes the child node indicated by oldChild from the list of children,
        and returns it. 
        """
        if not oldChild._id in self._child_ids:
            raise NotFoundException(oldChild)
        self._child_ids.remove(oldChild._id)
        del self._child_map[oldChild._id]
        return oldChild
            
    def appendChild(self, newChild):
        """
        Adds the node newChild to the end of the list of children of this node.
        If the newChild is already in the tree, it is first removed. 
        """
        if newChild.getparentNode():
            newChild.getparentNode().removeChild(newChild)
            newChild=newChild.aq_base
        self._child_ids.append(newChild._id)
        self._child_map[newChild._id]=newChild     
    
    # object manager methods
    # ----------------------

    def _objectIds(self, spec=None):
        # objectIds in native format
        if spec is not None:
            if type(spec)==type('s'):
                spec=[spec]
            return filter(lambda i, l=spec, m=self._child_map: m[i]._name in l,
                    self._child_ids)
        return self._child_ids

    def objectIds(self, spec=None):
        """
        Returns a list of sub-Element ids.
        """
        return map(lambda i, f=self._index2id: f(i), self._objectIds(spec))
        
    def objectValues(self, spec=None):
        """
        Returns a list of sub-Elements.
        """
        return map(lambda i, s=self, m=self._child_map: m[i].__of__(s),
                self._objectIds(spec))
    
    def objectItems(self, spec=None):
        """
        Returns a list of tuples (id, Element) of sub-Elements.
        """
        return map(lambda i, s=self, f=self._index2id, m=self._child_map:
                (f(i), m[i].__of__(s)), self._objectIds(spec))

    # misc Zope methods
    # -----------------

    def meta_type(self):
        # meta type is the node name
        return self.getnodeName()

    def tpValues(self):
        """
        Returns children of type Element.
        """
        return filter(lambda x: x.getnodeType()==ELEMENT_NODE, self.objectValues()) 
        
    def tpURL(self):
        """
        URL for tree navigation.
        """
        return self.id
        
    def text_content(self):
        """
        Concatination of all child Text nodes. Generally this only makes
        sense to call on leaf nodes.
        """
        result=""
        for child in self:
            if child.getnodeType()==TEXT_NODE:
                result=result + child.getnodeValue()
        return result

    def toXML(self, RESPONSE=None):
        """
        DOM tree as XML from this Node.
        """
        if RESPONSE is not None:
            RESPONSE.setHeader('content-type','text/xml')
        result=[]
        result.append('<%s' % self._name)
        for k,v in self.getattributes().items():
            if v:
                result.append(' %s="%s"' % (k,v))
        if self._child_ids:
            result.append('>')
            for child in self:
                result.append(child.toXML())
            result.append('</%s>' % self._name)
        else:
            result.append('/>')
        return string.join(result,'')

    document_src=toXML


class Text(Node):
    """Text Node"""
        
    nodeType=TEXT_NODE
    _name="#text"
    
    def __init__(self, data=""):
        self._data=data
        self._child_ids=()

    def toXML(self):
        """
        DOM tree as XML from this Node.
        """
        return self._data


    # DOM Methods
    # -----------
            
    def getdata(self):
        """
        The character data of this node.
        """
        return self._data
    
    getnodeValue=getdata
    
    def getlength(self):
        """
        The length character data of this node.
        """
        return len(self._data)
    
    def substringData(self,offset,count):
        """
        Extracts a range of data from the node.
        """
        return self.data[offset:offset+count]
        
    
class Element(Node):
    """Element Node"""

    def __init__(self, name, attributes=None):
        self._name=name
        self._child_ids=[]
        self._child_map={}
        if attributes is not None:
            for k, v in attributes.items():
                self.setAttribute(k,v)

    def getattributes(self):
        if not hasattr(self,'_attributes'):
            return {}
        return self._attributes

    # DOM Methods
    # -----------
    
    nodeType=ELEMENT_NODE
    
    def gettagName(self):
        return self._name
        
    def getAttribute(self, name):
        """
        Retrieves an attribute value by name.
        """
        try:
            return self._attributes[name]
        except:
            return ""

    def setAttribute(self, name, value):
        """
        Adds a new attribute. If an attribute with that name is already
        present in the element, its value is changed to be that of the
        value parameter.
        """
        if not hasattr(self,'_attributes'):
            self._attributes={}
        self._attributes[name]=value
        
    def getElementsByTagName(self,name):
        """
        Returns a Python list of all descendant elements with a given tag
        name, in the order in which they would be encountered in a preorder
        traversal of the Element tree. 
        """
        result=[]
        for child in self:
            if child.nodeType==ELEMENT_NODE and child.gettagName()==name or \
                    name=="*":
                result.append(child.__of__(self))
                result=result + child.__of__(self).getElementsByTagName(name)
        return result


    # Object Manager-ish extensions
    # -----------------------------
    # Hmm. Are these worth it?
    
    def nextObject(self, spec=None):
        """
        Returns the next sibling element of type spec.
        """
        e=self
        if type(spec)==type(''):
            spec=[spec]
        while 1:
            e=e.getnextSibling()
            if e is None:
                return None
            if spec is None or e._name in spec:
                return e
        
    def previousObject(self, spec=None):
        """
        Returns the previous sibling element of type spec.
        """
        e=self
        if type(spec)==type(''):
            spec=[spec]
        while 1:
            e=e.getpreviousSibling()
            if e is None:
                return None
            if spec is None or e._name in spec:
                return e


class Document(Node):
    """Document Node"""
    
    nodeType=DOCUMENT_NODE
    docType=None
    _name="#document"
    
    def __init__(self, attributes=None):
        self._child_map={}
        self._child_ids=[]
        self._next_id=0

    def toXML(self,RESPONSE=None):
        """
        DOM tree as XML from this Node.
        """
        if RESPONSE is not None:
            RESPONSE.setHeader('content-type','text/xml')
        result = '<?xml version="1.0"?>\n'
        for child in self:
            result=result + child.toXML()
        return result

    # DOM Methods
    # -----------

    def getdocType(self):
        """
        The Document Type Declaration associated with this document.
        """
        return self.docType
        
    def getdocumentElement(self):
        """
        The child node that is the root element of the document.
        """
        for child in self:
            if child.getnodeType() == ELEMENT_NODE:
                return child
    
    def getElementsByTagName(self,name):
        """
        Returns a Python list of all descendant elements with a given tag
        name, in the order in which they would be encountered in a preorder
        traversal of the Element tree. 
        """
        return self.getdocumentElement().getElementsByTagName(name)


        
class ProcessingInstruction(Node):
    """Processing Instruction"""
    
    nodeType=PROCESSING_INSTRUCTION_NODE
    
    def __init__(self, target, data):
        self._target=target
        self._name=target
        self._data=data
        self._child_ids=()
        
    def toXML(self):
        """
        DOM tree as XML from this Node.
        """
        return "<?%s %s?>" % (self._target, self._data)
    
